using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using SimpleJSON;

namespace EmissiveClothing
{
    public interface IEmissiveSettings
    {
        bool DisableEmissive { get; }
        bool IsSession { get; }
        bool AutomaticTextureSettings { get; }
        float Alpha { get; }
        float AlphaMultiplier { get; }
        bool RenderOriginal { get; }
        bool RenderOriginalWithoutDecal { get; }
        string TexPrioritizeOrder { get; }
        bool ShaderMode { get; }
        bool OriginalColor { get; }
        bool ColorMode { get; }
        Color Color { get; }
        bool LiveReload { get; }
        bool Verbose { get; }
        bool FilterActive { get; }
        IEnumerable<string> FilterNames { get; }
        Shader EmissiveShader { get; }
        bool InitScriptDone { get; }
    }

    public class EmissiveClothingPlus : MVRScript, IEmissiveSettings
    {

        protected JSONStorableBool _disableEmissive;
        protected JSONStorableBool _automaticSettingsFromTextureSource;

        protected JSONStorableFloat _alpha;
        protected JSONStorableFloat _alphaMultiplier;
        protected JSONStorableBool _renderOriginal;
        protected JSONStorableBool _renderOriginalWithoutDecal;
        protected JSONStorableStringChooser _texPrioritizeOrder;
        protected JSONStorableBool _shaderMode;
        // protected JSONStorableUrl _loadedShaderPath = new JSONStorableUrl("shader", "");
		protected JSONStorableBool _originalColor;
		protected JSONStorableBool _colorMode;
        protected JSONStorableColor _color;

        protected JSONStorableBool _liveReload;
        protected JSONStorableBool _verbose;
        protected JSONStorableBool _filterActive;
        protected JSONStorableString _filterStore;
        protected List<JSONStorableBool> _filterToggles;
        protected List<string> _filterNames;

        private bool _isSession = false;

        public bool DisableEmissive => _disableEmissive.val;
        public bool IsSession => _isSession;
        public bool AutomaticTextureSettings => _automaticSettingsFromTextureSource.val;
        public float Alpha => _alpha.val;
        public float AlphaMultiplier => _alphaMultiplier.val;
        public bool RenderOriginal => _renderOriginal.val;
        public bool RenderOriginalWithoutDecal => _renderOriginalWithoutDecal.val;
        public string TexPrioritizeOrder => _texPrioritizeOrder.val;
        public bool ShaderMode => _shaderMode.val;
        public bool OriginalColor => _originalColor.val;
        public bool ColorMode => _colorMode.val;
        public Color Color => _color.colorPicker.currentColor;
        public bool LiveReload => _liveReload.val;
        public bool Verbose => _verbose.val;
        public bool FilterActive => _filterActive.val;
        public IEnumerable<string> FilterNames => _filterNames;
        public Shader EmissiveShader => _emissiveShader;
        public bool InitScriptDone => _initScriptDone;

        private static readonly string BUNDLE_NAME = "emissiveshader.assetbundle";        

        
        protected List<AtomWrapper> _atomWrappers;

        protected Shader _emissiveShader;
        
        protected bool _initScriptDone = false;
        protected bool _filterDirty = false;

        const string TEX_SOURCE_DECALTEX = "_DecalTex > _AlphaTex";
        const string TEX_SOURCE_ALPHATEX = "_AlphaTex > _DecalTex";
        const string TEX_SOURCE_SPECTEX = "_SpecTex > _DecalTex";
        const string TEX_SOURCE_GLOSSTEX = "_GlossTex > _DecalTex";
        const string TEX_SOURCE_BUMPMAP = "_BumpMap > _DecalTex";
        const string TEX_SOURCE_MAINTEX = "_MainTex > _DecalTex";

        public override void Init()
        {
            try
            {

                _isSession = GetIsSession();

                _atomWrappers = new List<AtomWrapper>();
                if (!IsSession)
                {
                    _atomWrappers.Add(new AtomWrapper(this, containingAtom, (IEmissiveSettings)this));
                }

                _initScriptDone = false;

                _filterNames = new List<string>();
                _filterToggles = new List<JSONStorableBool>();

                _color = new JSONStorableColor("Color", HSVColorPicker.RGBToHSV(1f, 1f, 1f), (JSONStorableColor _) => SyncMatsAll());
                RegisterColor(_color);
                CreateColorPicker(_color, true).label = "Color";

                _alpha = new JSONStorableFloat("ColorAlpha", 1f, (float _) => { SyncMatsAll(); }, 0f, 4f, false, true);
                RegisterFloat(_alpha);
                CreateSlider(_alpha, true).label = "Color Alpha";
                _alpha.valNoCallback = 1.2f;

                _alphaMultiplier = new JSONStorableFloat("ColorAlphaMultiplier", 1f, (float _) => { SyncMatsAll(); }, 0f, 4f, false, true);
                RegisterFloat(_alphaMultiplier);
                var slider = CreateSlider(_alphaMultiplier, true);
                slider.label = "Alpha Multiplier";

                _disableEmissive = new JSONStorableBool("DisableEmissive", false, (bool _) => StartRebuildAll());
                RegisterBool(_disableEmissive);
                var toggle = CreateToggle(_disableEmissive, false);
                toggle.label = "Disable Emissive";
                toggle.backgroundColor = new Color(0.9f, 0.5f, 0.5f);

                _automaticSettingsFromTextureSource = new JSONStorableBool("AutomaticSettingsFromTextureSource", true, (bool _) => StartRebuildAll());
                RegisterBool(_automaticSettingsFromTextureSource);
                toggle = CreateToggle(_automaticSettingsFromTextureSource, false);
                toggle.label = "Automatic Settings\nfrom decal texture source name where applicable*";
                toggle.backgroundColor = new Color(0.8f, 0.9f, 0.8f);
                toggle.height = 120;

                CreateSpacer(false).height = 30;

                _renderOriginal = new JSONStorableBool("RenderOriginalMaterial", true, (bool _) => StartRebuildAll());
                RegisterBool(_renderOriginal);
                CreateToggle(_renderOriginal, false).label = "Render Original Material";

                _shaderMode = new JSONStorableBool("ShaderEmissiveMode", false, (bool _) => StartRebuildAll());
                RegisterBool(_shaderMode);
                toggle = CreateToggle(_shaderMode, false);
                toggle.label = "Shader Emissive Mode (Sim)";
                // toggle.backgroundColor = new Color(0.9f, 0.9f, 0.7f);

                _originalColor = new JSONStorableBool("UseOriginalColor", true, (bool _) => SyncMatsAll());
                RegisterBool(_originalColor);
                CreateToggle(_originalColor, false).label = "Use Original Color";

                // premultiplied IsSession standard, false for legacy compatability with older plugins how it worked there!
                _colorMode = new JSONStorableBool("ColorMode", IsSession, (bool _) => SyncMatsAll());
                RegisterBool(_colorMode);
                toggle = CreateToggle(_colorMode, false);
                toggle.label = "Color Mode (Premultiplied)";

                _renderOriginalWithoutDecal = new JSONStorableBool("RemoveDecalFromSource", false, (bool _) => StartRebuildAll());
                RegisterBool(_renderOriginalWithoutDecal);
                toggle = CreateToggle(_renderOriginalWithoutDecal, false);
                toggle.label = "Remove Decal texture from source when 'Render Original' is active";
                toggle.height = 80;
                toggle.backgroundColor = new Color(0.9f, 0.7f, 0.7f);

                List<string> pulseMainTypes = new List<string>();
                pulseMainTypes.Add(TEX_SOURCE_DECALTEX);
                pulseMainTypes.Add(TEX_SOURCE_ALPHATEX);
                pulseMainTypes.Add(TEX_SOURCE_SPECTEX);
                pulseMainTypes.Add(TEX_SOURCE_GLOSSTEX);
                pulseMainTypes.Add(TEX_SOURCE_BUMPMAP);
                pulseMainTypes.Add(TEX_SOURCE_MAINTEX);
                _texPrioritizeOrder = new JSONStorableStringChooser("TexSourceOrder", pulseMainTypes, TEX_SOURCE_DECALTEX, "TexSourceOrder", (string type) => { StartRebuildAll(); });
                RegisterStringChooser(_texPrioritizeOrder);
                var popup = CreateScrollablePopup(_texPrioritizeOrder, false);
                popup.label = "Texture source priority";

                // RegisterUrl(_loadedShaderPath);

                CreateSpacer(false).height = 30;

                _verbose = new JSONStorableBool("Verbose", false, (bool _) => { });
                RegisterBool(_verbose);
                toggle = CreateToggle(_verbose, false);
                toggle.label = "Verbose";
                // toggle.height = 80;
                toggle.backgroundColor = new Color(0.65f, 0.85f, 0.9f);

                if (!IsSession)
                {
                    var btnb = CreateButton("Output Current Settings", false);
                    btnb.button.onClick.AddListener(() => OutputCurrentSettings());
                    btnb.buttonColor = new Color(0.65f, 0.85f, 0.9f);
                    // btn.height = 80;
                }

                _liveReload = new JSONStorableBool("LiveRescan", true, (bool _) => { });
                RegisterBool(_liveReload);
                toggle = CreateToggle(_liveReload, false);
                toggle.label = "Live Rescan after emissive clothing changes";
                toggle.height = 80;
                toggle.backgroundColor = new Color(0.8f, 0.9f, 0.8f);

                var btn = CreateButton("Rescan active clothes");
                btn.button.onClick.AddListener(() => StartRebuildAll());
                btn.buttonColor = new Color(0.9f, 0.9f, 0.7f);
                btn.height = 80;

                // action callable by triggers if needed to rescan clothes
                RegisterAction(new JSONStorableAction("RescanActiveClothes", () =>
                {
                    StartRebuildAll();
                }));

                _filterActive = new JSONStorableBool("FilterOverride", false, (bool _) =>
                {
                    UpdateFilterUI();
                    SyncFilter();
                });
                RegisterBool(_filterActive);
                if (!IsSession)
                {
                    toggle = CreateToggle(_filterActive, false);
                    toggle.label = "Filter Override";
                    toggle.backgroundColor = new Color(0.9f, 0.9f, 0.7f);
                }

                _filterStore = new JSONStorableString("emissiveFilterNames", "");
                if (!IsSession)
                {
                    RegisterString(_filterStore);
                }

                System.Text.StringBuilder sb = new System.Text.StringBuilder();
                if (IsSession)
                {
                    sb.AppendLine("<b>Session mode active:</b>");
                    sb.AppendLine("Track person atoms without EmissiveClothingPlus scripts in the scene after emissive clothing. 'Filter Override' and 'Output Current Settings' are disabled.");
                    sb.AppendLine("You need to rescan if a look and clothing are already loaded in-memory when used again.");
                    sb.AppendLine("");
                }
                sb.AppendLine("<b>Disable Emissive:</b>");
                sb.AppendLine("Disables active emissive effects and prevents the new emissive effects from activating from this instance.");
                sb.AppendLine("");
                sb.AppendLine("<b>*Automatic Settings:</b>");
                sb.AppendLine("Prioritize to use emissive settings parsed from the decal texture source name if available while 'Filter Override' is disabled. Only color setting is available to change (if 'Use Original Color' is unchecked) in this mode.");
                sb.AppendLine("");
                sb.AppendLine("<b>Render Original Material:</b>");
                sb.AppendLine("Renders the original material together with the emissive material. Different results when emissive effect is low/transparent.");
                sb.AppendLine("");
                sb.AppendLine("<b>Shader Emissive Mode (Sim):</b>");
                sb.AppendLine("Off: Recommended standard mode. Renders the emissive effect using a duplicate material to not modify built in materials in VAM. Some clothing features will not work, such as sim enabled clothing.");
                sb.AppendLine("On: The emissive effect will be available with all clothing features, such as sim enabled clothing. Renders the emissive effect by swapping the clothing materials with the emissive one. In theory, this mode breaks if other scripts also modify the material. Recommended to only use this mode when you need sim enabled clothing!");
                sb.AppendLine("");
                sb.AppendLine("<b>Use Original Color:</b>");
                sb.AppendLine("On: Use color from the material's diffuse color.");
                sb.AppendLine("Off: Uses color from the plugin.");
                sb.AppendLine("Color tints the emissive texture source! Depending on the source texture, recoloring might not work as you expect.");
                sb.AppendLine("");
                sb.AppendLine("<b>Color Mode (Premultiplied):</b>");
                sb.AppendLine("Toggles premultiplied coloring on/off when syncing materials.");
                sb.AppendLine("");
                sb.AppendLine("<b>Color Alpha:</b>");
                sb.AppendLine("The base strength of the emissive effect which can be > 1. This value is multiplied with the alpha multiplier.");
                sb.AppendLine("");
                sb.AppendLine("<b>Alpha Multiplier:</b>");
                sb.AppendLine("This value is multiplied with the color alpha value to produce the final alpha value. The alpha multiplier is intended to be updated through other triggers and scripts!");
                sb.AppendLine("");
                sb.AppendLine("<b>Remove Decal texture from source when 'Render Original' is active:</b>");
                sb.AppendLine("If the decal from the source material should be cleared when rendering the original material. Has a greater visual impact when the emissive effect is low.");
                sb.AppendLine("");
                sb.AppendLine("<b>Texture source priority:</b>");
                sb.AppendLine("Default texture used for the emissive shader is the _DecalTex (for purpose built emissive clothing items). Prioritize to use another texture source for the emissive shader here.");
                sb.AppendLine("");
                sb.AppendLine("<b>Verbose:</b>");
                sb.AppendLine("When on, log messages about certain key events, status messages and warnings in this script.");
                sb.AppendLine("Turn off to silence this script for a better user experience.");
                sb.AppendLine("");
                if (!IsSession)
                {
                    sb.AppendLine("<b>Output Current Settings:</b>");
                    sb.AppendLine("Intended for creators to grab emissive settings as a parsable string to include in an emissive decal texture name for automatic settings.");
                    sb.AppendLine("");
                }
                sb.AppendLine("<b>Live Rescan after emissive clothing changes:</b>");
                sb.AppendLine("After loading detected (which usually happens when loading clothing items or presets), any emissive clothing item that has changed will trigger 'Rescan active clothes' to be run automatically.");
                sb.AppendLine("");
                sb.AppendLine("<b>Rescan active clothes:</b>");
                sb.AppendLine("Perform a manual rescan and emissive rebuild. Recommended to do after force-reload or clothing preset change on an active emissive clothing item.");
                sb.AppendLine("");
                if (!IsSession)
                {
                    sb.AppendLine("<b>Filter Override:</b>");
                    sb.AppendLine("Clothing materials with '(em)' in the name are automatically emissive. Enable 'Filter Override' to manually select clothing materials to be emissive. Emissive effect on non-emissive clothing materials varies depending on the selected source texture.");
                    sb.AppendLine("");
                    sb.AppendLine("<b>Other:</b>");
                    sb.AppendLine("You can have multiple instances of this plugin on the same person but only one emissive effect will be active per clothing item!");
                }

                JSONStorableString infoJson = new JSONStorableString("infoText", sb.ToString());
                UIDynamicTextField textField = CreateTextField(infoJson, true);
                SetTextFieldHeight(textField, 540);

                StartCoroutine(LoadShaderAndInit());
            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        #region PLUGIN_UI
        public void UpdateFilterUI()
        {
            // Read wraps and create filter

            // remove old UI
            foreach (JSONStorableBool b in _filterToggles)
            {
                RemoveToggle(b);
            }
            _filterToggles.Clear();

            if (!_filterActive.val || IsSession)
            {
                return;
            }

            DAZSkinWrap[] allWraps = containingAtom.gameObject.GetComponentsInChildren<DAZSkinWrap>(false);

            List<FilterOption> availableFilters = new List<FilterOption>();
            foreach (DAZSkinWrap wrap in allWraps)
            {
                // SuperController.LogMessage("wrap=" + wrap.ToString());
                if (wrap.ToString().Contains("(" + nameof(EmissiveDAZSkinWrap) + ")"))
                {
                    continue;
                }
                if (wrap.skin.delayDisplayOneFrame)
                {
                    continue;
                }

                bool hasMaterialSlots = wrap.GetComponents<DAZSkinWrapMaterialOptions>().Any(mo => mo.paramMaterialSlots?.Length > 0);
                if (!hasMaterialSlots)
                {
                    continue;
                }

                int materialOption = 0;
                foreach (DAZSkinWrapMaterialOptions option in wrap.GetComponents<DAZSkinWrapMaterialOptions>())
                {
                    if (option.paramMaterialSlots?.Length > 0)
                    {
                        //if (IsEmissiveFlagged(option))
                        //{
                        //    // a purpose built emissive clothing item
                        //}

                        string key = wrap.ToString() + "_" + wrap.wrapName + "_" + option.storeId;
                        FilterOption filterOption = new FilterOption(key, GetClothingName(option.storeId), materialOption, FancifyMaterialOptionLabel(wrap.wrapName, option.storeId));
                        availableFilters.Add(filterOption);

                        // SuperController.LogMessage(key);

                    }
                    materialOption++;
                }

            }

            availableFilters.Sort((a, b) => {
                int r = string.Compare(a.Name, b.Name);
                if (r == 0)
                {
                    if (a.MaterialSlot == b.MaterialSlot)
                        return 0;
                    else
                        return (a.MaterialSlot < b.MaterialSlot ? 1 : -1);
                }

                return r;
            });

            // SuperController.LogMessage("Available filters: " + string.Join(", ", availableFiltersEm.ToArray()));
            ReadFilter();

            int bgC = 0;
            string lastName = "";

            foreach (FilterOption option in availableFilters)
            {
                if (option.Key == null || option.Key == "")
                    continue;

                if (lastName != option.Name)
                {
                    bgC++;
                    lastName = option.Name;
                }

                bool initialSelected = _filterNames.Contains(option.Key);
                JSONStorableBool toggle = new JSONStorableBool(option.Key, initialSelected, (bool _) => SyncFilter());
                var tg = CreateToggle(toggle, false);
                tg.height = 120;
                tg.backgroundColor = (bgC % 2 == 0 ? new Color(0.8f, 0.9f, 0.9f) : new Color(0.8f, 0.85f, 0.9f));
                tg.label = option.Label;

                //if (tg.label.Contains("(em)"))
                //{
                //    // a purpose built emissive clothing item
                //    tg.backgroundColor = (bgC % 2 == 0 ? new Color(0.9f, 0.9f, 0.8f) : new Color(0.9f, 0.9f, 0.9f));
                //}

                _filterToggles.Add(toggle);
            }
        }
        #endregion

        #region SHADER_LOADER
        // MeshVR.AssetLoader provides some nice refcounting for us.
        // The downsides are the annoying callback API, that we can't suppress error messages, and that a refcount really shouldn't be created on failure, wat
        class LoadAssetBundle : CustomYieldInstruction
        {
            AssetLoader.AssetBundleFromFileRequest req;

            private bool done;

            public bool success => req.assetBundle != null;
            public T GetAsset<T>(string name) where T : UnityEngine.Object
            {
                return req.assetBundle?.LoadAsset<T>(name);
            }

            public LoadAssetBundle(string url)
            {
                req = new AssetLoader.AssetBundleFromFileRequest
                {
                    path = url,
                    callback = _ => {
                        done = true;
                        if (req.assetBundle == null)
                            AssetLoader.DoneWithAssetBundleFromFile(req.path);
                    }
                };
                AssetLoader.QueueLoadAssetBundleFromFile(req);
            }

            public override bool keepWaiting => !done;
        }

        //public static bool IsErrorLogOpen()
        //{
        //    if (SuperController.singleton.errorLogPanel != null && SuperController.singleton.errorLogPanel.gameObject != null)
        //    {
        //        return SuperController.singleton.errorLogPanel.gameObject.activeSelf;
        //    }
        //    return false;
        //}

        private IEnumerator AttemptLoadShader(string url)
        {
            if (_emissiveShader != null)
                yield break;

            // bool errorLogAlreadyOpen = IsErrorLogOpen();

            {
                Shader shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
                if (shader != null)
                {
                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: Found already loaded emissiveshader assetbundle with same shader name, trying to use it.");

                    _emissiveShader = shader;
                    yield break;
                }
            }

            var request = new LoadAssetBundle(url);
            yield return StartCoroutine(request);

            if (!request.success)
            {
                if (_emissiveShader == null)
                {
                    Shader shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
                    if (shader != null)
                    {
                        if (_verbose.val)
                            SuperController.LogMessage("EmissiveClothing: Found already loaded emissiveshader assetbundle with same shader name, trying to use it.");

                        //// a shader instance already exist, try to use it
                        //if (!errorLogAlreadyOpen)
                        //    SuperController.singleton.CloseErrorLogPanel();

                        _emissiveShader = shader;
                    }
                }
                yield break;
            }

            try
            {
                _emissiveShader = request.GetAsset<Shader>("Assets/EmissiveHack.shader");
                if (_emissiveShader != null)
                {
                    // _loadedShaderPath.val = url;
                }
                else
                {
                    SuperController.LogError("Bad emissiveshader assetbundle");
                    try
                    {
                        AssetLoader.DoneWithAssetBundleFromFile(url);
                    }
                    catch
                    {
                    }
                }
            }
            catch (Exception e)
            {
                if (_verbose.val)
                    SuperController.LogMessage("EmissiveClothing: Failed to load emissiveshader assetbundle: " + e.ToString());
            }

            if (_emissiveShader == null)
            {
                Shader shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
                if (shader != null)
                {
                    if (_verbose.val)
                        SuperController.LogMessage("EmissiveClothing: Found already loaded emissiveshader assetbundle with same shader name, trying to use it.");

                    //// a shader instance already exist, try to use it
                    //if (!errorLogAlreadyOpen)
                    //    SuperController.singleton.CloseErrorLogPanel();                    

                    _emissiveShader = shader;
                }
            }
        }

        // If there are multiple copies of this script loaded from different paths, avoid failures due to trying to load the same assetbundle. 
        // This is particularly important because sometimes the bundle seems to get leaked and not unload the shader.
        // Stash the url as well so that vac saves work properly even if this happens
        /* private static readonly string STASH_NAME = "Alazi.EmissiveShaderStash";
        private void StashShaderName()
        {
            PlayerPrefs.SetString(STASH_NAME, loadedShaderPath.val);
        }
        private void LoadStashedShader()
        {
            shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
            if (shader != null) {
                loadedShaderPath.val = PlayerPrefs.GetString(STASH_NAME);
            } else {
                PlayerPrefs.DeleteKey(STASH_NAME);
            }
        } */

        protected IEnumerator LoadShaderAndInit()
        {
            // SuperController.LogMessage("LoadShaderAndInit");
            // Wait until json load is done
            yield return new WaitForEndOfFrame();
            //LoadStashedShader();
            //SuperController.LogMessage("LoadStashedShader");
            if (_emissiveShader == null)
            {

                string pluginBundlePath = GetPluginBundlePath(this);

                if (_verbose.val)
                    SuperController.LogMessage("EmissiveClothing: Shader path: " + pluginBundlePath);
                // Loading the same asset from exactly the same source is OK
                yield return AttemptLoadShader(pluginBundlePath);

                if (_emissiveShader == null)
                {
                    SuperController.LogError("EmissiveClothing: Failed to load shader");
                    yield break;
                }
            }

            yield return WaitLoading(_verbose.val);
            yield return new WaitForEndOfFrame();

            // let emissive plugins with filters activate FIRST by waiting a second longer after loading, let other threads run first
            // This is to avoid collisions when multiple emissive scripts overlap. This is not a supported scenario.
            if (!_filterActive.val)
            {
                yield return new WaitForSeconds(1f);
                yield return new WaitForEndOfFrame();
                yield return new WaitForEndOfFrame();
            }

            yield return new WaitForEndOfFrame();
            UpdateFilterUI();

            _restoreFromJSONDone = true;
            _initScriptDone = true;

            if (IsSession)
            {
                yield return new WaitForSeconds(1f);
                yield return new WaitForEndOfFrame();
                RescanWrappers();
            }
            else
            {
                WrappersAction((x) => x.StartRebuild());
            }
        }
        #endregion

        #region PLUGIN_SESSION_CORE
        public bool GetIsSession()
        {
            Atom a = this.GetContainingAtom();
            if (a != null && (a.type == "SessionPluginManager")) // || a.type == "CoreControl"))
            {
                return true;
            }
            return false;
        }

        private bool _restoreFromJSONDone = false;
        private bool _restoredFromJSON = false;
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            // Ensure restore from JSON is only done once and not by misstake later when load-merge another plugin.
            // If the load-merge plugin matches this one and this has not yet run its RestoreFromJSON, the load-merge
            // will call RestoreFromJSON on 2 plugin instances. This one and the newly added from the load-merge.
            if (!_restoreFromJSONDone)
            {
                _restoreFromJSONDone = true;
                _restoredFromJSON = true;
                base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
                ReadFilter();
            }
        }

        protected void SyncFilter()
        {
            // forbidden as long as script is initing or rebuild is already in action
            if (!_initScriptDone || Wrappers.Any((x) => x.Rebuilding))
                return;

            if (_verbose.val)
                SuperController.LogMessage("EmissiveClothing: Syncing filters");

            if (_filterActive.val)
            {
                string filter = string.Join(";", _filterToggles.Where(x => x.val).Select(x => x.name).ToArray());
                _filterStore.val = filter;
                _filterDirty = true;
            }
            ReadFilter();

            StartRebuildAll();
        }

        protected void ReadFilter()
        {
            string filter = _filterStore.val;
            _filterNames.Clear();
            _filterNames.AddRange(filter.Split(new char[] { ';' }));
            if (filter != "" && _filterDirty)
            {
                // only log when filter is dirty to avoid "spam" as the filter is updated through different paths in the code
                if (_verbose.val)
                {
                    SuperController.LogMessage("EmissiveClothing: Filters");
                    foreach (string s in _filterNames)
                    {
                        SuperController.LogMessage(s);
                    }
                }
            }
            _filterDirty = false;
        }

        class FilterOption
        {
            public string Key;
            public string Name;
            public int MaterialSlot;
            public string Label;
            public FilterOption(string key, string name, int materialSlot, string label)
            {
                Key = key;
                Name = name;
                MaterialSlot = materialSlot;
                Label = label;
            }
        }

        protected void OutputCurrentSettings()
        {
            SuperController.LogMessage("EmissiveClothing:");
            SuperController.LogMessage("This outputs all active emissive materials as a parsable string to include in decal texture names for automatic settings to apply.");
            SuperController.LogMessage("Automatic settings only works when applied to decal texture names using a decal texture source as emissive!");

            if (_automaticSettingsFromTextureSource.val)
            {
                SuperController.LogMessage("");
                SuperController.LogMessage("To output active emissive materials, first disable 'Automatic Settings' and set emissive settings in the plugin.");
                return;
            }

            int count = 0;
            WrappersAction((x) => {
                if (x.OutputCurrentSettings())
                {
                    count++;
                }
            });
            SuperController.LogMessage("--------------------------------------------------------------------------------------------------------------------------------------------------------------");
            if (count == 0)
            {
                SuperController.LogMessage("No active emissive materials found to output.");
            }
            else
            {
                SuperController.LogMessage("Total output: " + count.ToString() + " item wraps.");
            }
            SuperController.singleton.OpenMessageLogPanel();
        }

        string _sessionLastLoadedScene = "";
        bool _sessionRedoWrappers = false;
        int _sessionUpdateWrappers = 0;
        void Update()
        {
            if (IsSession)
            {
                if (_sessionLastLoadedScene != SuperController.singleton.LoadedSceneName)
                {
                    _sessionLastLoadedScene = SuperController.singleton.LoadedSceneName;
                    _sessionRedoWrappers = true;
                    _sessionUpdateWrappers = 20;
                } 
                else if (IsVAMLoading())
                {    
                    _sessionUpdateWrappers = Math.Max(3, _sessionUpdateWrappers);
                }
                else if (_sessionUpdateWrappers > 0 && _initScriptDone)
                {
                    _sessionUpdateWrappers--;
                    if (_sessionUpdateWrappers <= 0)
                    {
                        if (_sessionRedoWrappers)
                        {
                            DestroyWrappers();
                            _sessionRedoWrappers = false;
                        }
                        RescanWrappers();
                    }
                }
                if (_sessionRedoWrappers)
                {
                    return;
                }
            }

            WrappersAction((x) => x.Update());
        }

        void OnDestroy()
        {
            WrappersAction((x) => x.Unbuild());
        }
                
        protected IEnumerable<AtomWrapper> Wrappers { get { return _atomWrappers; } }

        protected void WrappersAction(Action<AtomWrapper> callback)
        {
            foreach(AtomWrapper wrapper in Wrappers)
            {
                callback.Invoke(wrapper);
            }
        }

        public void SyncMatsAll()
        {
            WrappersAction((x) => x.SyncMats());
        }

        public void StartRebuildAll()
        {
            if (IsSession)
            {
                RescanWrappers();
            }
            WrappersAction((x) => x.StartRebuild());
            StartCoroutine(UpdateFilterUIAfterRebuildDone());
        }
        protected IEnumerator UpdateFilterUIAfterRebuildDone()
        {
            while (Wrappers.Any((x) => x.Rebuilding))
            {
                yield return new WaitForSeconds(0.2f);
                yield return new WaitForEndOfFrame();
            }
            UpdateFilterUI();
        }

        public void RescanWrappers()
        {
            // remove wrapper whos atoms are no longer active
            bool change = false;
            HashSet<Atom> activeWrapperAtoms = new HashSet<Atom>();
            for (int i = 0; i < _atomWrappers.Count; i++)
            {
                if (_atomWrappers[i].containingAtom.destroyed)
                {
                    _atomWrappers[i].Unbuild();
                    _atomWrappers.RemoveAt(i);
                    i--;
                    change = true;
                } 
                else
                {
                    activeWrapperAtoms.Add(_atomWrappers[i].containingAtom);
                }
            }
            // create wrappers for atoms that's active
            List<AtomWrapper> newAtomWrappers = null;
            foreach (Atom atom in SuperController.singleton.GetAtoms().Where((x) => { return x.type == "Person"; }))
            {
                if (activeWrapperAtoms.Contains(atom))
                    continue;

                AtomWrapper atomWrapper = new AtomWrapper(this, atom, (IEmissiveSettings)this);
                _atomWrappers.Add(atomWrapper);
                activeWrapperAtoms.Add(atom);
                if (newAtomWrappers == null)
                    newAtomWrappers = new List<AtomWrapper>();
                newAtomWrappers.Add(atomWrapper);
                change = true;
            }
            if (_verbose.val && change)
            {
                SuperController.LogMessage("EmissiveClothing: Persons tracked (from session instance): " + string.Join(", ", _atomWrappers.Select((x) => x.containingAtom.uid).ToArray()));
            }
            if (newAtomWrappers != null)
            {
                foreach(AtomWrapper newAtomWrapper in newAtomWrappers)
                {
                    newAtomWrapper.StartRebuild(0.3f);
                }
            }
        }

        public void DestroyWrappers()
        {
            WrappersAction((x) => x.Unbuild());
            _atomWrappers.Clear();
        }
        #endregion

        #region ATOM_WRAPPER
        public class AtomWrapper
        {
            protected List<EmissiveDAZSkinWrap> _ewraps = new List<EmissiveDAZSkinWrap>();
            protected List<DAZSkinWrap> _wraps = new List<DAZSkinWrap>();
            protected List<EmissiveMaterialOptionData> _emissiveMaterialData;

            protected volatile bool _rebuilding = false;
            public bool Rebuilding { get { return _rebuilding; } }

            public Atom containingAtom;
            public IEmissiveSettings settings;

            protected string _texPrioritizeOrder => settings.TexPrioritizeOrder;
            protected bool _liveReload => settings.LiveReload;
            protected bool _verbose => settings.Verbose;
            protected bool _filterActive => settings.FilterActive;
            protected IEnumerable<string> _filterNames => settings.FilterNames;
            protected Shader _emissiveShader => settings.EmissiveShader;
            protected bool _initScriptDone => settings.InitScriptDone;

            protected bool _waitForOn = false;

            protected MVRScript _masterScript;

            public AtomWrapper(MVRScript masterScript, Atom atom, IEmissiveSettings emissiveSettings)
            {
                _masterScript = masterScript;
                containingAtom = atom;
                settings = emissiveSettings;
                _emissiveMaterialData = new List<EmissiveMaterialOptionData>();
            }

            protected int _queueLiveRescan = 0;
            public void Update()
            {
                if (_emissiveMaterialData == null)
                    return;

                if (_waitForOn)
                {
                    if (!_initScriptDone || _rebuilding)
                        return;

                    if (containingAtom != null && containingAtom.on)
                    {
                        if (_verbose)
                            SuperController.LogMessage("EmissiveClothing: Atom " + containingAtom.uid + " turned on, rebuilding");

                        _waitForOn = false;
                        StartRebuild(0.1f);
                    }
                    return;
                }

                SyncMats();

                if (_initScriptDone && !_rebuilding && !settings.DisableEmissive)
                {
                    //foreach (DAZSkinWrap wrap in _wraps)
                    //{ wrap.isActiveAndEnabled }

                    if (IsVAMLoading())
                    {
                        _queueLiveRescan = (settings.IsSession ? 6 : 3); // wait double updates in session to allow non-session plugins to get ahead
                    }
                    else if (_queueLiveRescan > 0)
                    {
                        if (!_liveReload)
                        {
                            _queueLiveRescan = 0;
                            return;
                        }

                        _queueLiveRescan --;
                        if (_queueLiveRescan <= 0)
                        {
                            if (_verbose)
                                SuperController.LogMessage("EmissiveClothing: (Live Rescan) Performing check for " + containingAtom.uid);

                            bool rebuild = false;

                            for (int i = 0; i < _wraps.Count; i++)
                            {
                                if (CheckWrapMaterialsChanged(_wraps[i], _ewraps[i]))
                                {
                                    if (_verbose)
                                        SuperController.LogMessage("EmissiveClothing: Wraps changed, performing rebuild");

                                    rebuild = true;
                                    break;
                                }
                            }

                            if (!rebuild && !_filterActive)
                            {
                                rebuild = ExistEMWrapWithoutEmissiveActive(containingAtom, _verbose);
                            }

                            if (rebuild)
                            {
                                StartRebuild(0.3f);
                            }
                        }
                    }
                }
            }

            public void StartRebuild(float delaySeconds = 0f)
            {
                if (!_initScriptDone || _rebuilding)
                    return;

                _rebuilding = true;
                _masterScript.StartCoroutine(Rebuild(delaySeconds));
            }

            protected IEnumerator Rebuild(float delaySeconds = 0f)
            {
                if (_verbose)
                    SuperController.LogMessage("EmissiveClothing: Rebuild emissive for " + containingAtom.uid);

                if (delaySeconds > 0f)
                {
                    yield return new WaitForSeconds(delaySeconds);
                    yield return new WaitForEndOfFrame();
                }

                Unbuild();

                if (settings.DisableEmissive)
                {
                    if (_verbose)
                        SuperController.LogMessage("EmissiveClothing: Ignoring build for " + containingAtom.uid + ". Emissive disabled.");

                    _rebuilding = false;
                    yield break;
                }

                // wait for components to be destroyed
                yield return new WaitForEndOfFrame();

                yield return WaitLoading(_verbose);
                yield return new WaitForEndOfFrame();

                if (settings.IsSession)
                {
                    // session mode will wait longer for non-session emissive plugins to get priority to build emissive stuff during livescan
                    yield return new WaitForSeconds(1f);
                    // a few end of frames too, to let plugin on new atoms load in properly - in case there was a freeze/lag spike on 1+ second no real frames have passed.
                    yield return new WaitForEndOfFrame();
                    yield return new WaitForEndOfFrame();
                    yield return new WaitForEndOfFrame();
                    yield return new WaitForEndOfFrame();
                    yield return new WaitForEndOfFrame();

                    if (ContainsScript(containingAtom, "EmissiveClothing.EmissiveClothingPlus"))
                    {
                        // the emissive plugin on the atom is responsible for building everything!
                        if (_verbose)
                            SuperController.LogMessage("EmissiveClothing: Ignoring build for '" + containingAtom.uid + "' (from session instance). EmissiveClothingPlus script already exist on atom.");

                        _rebuilding = false;
                        yield break;
                    }
                }

                Build();

                _rebuilding = false;
            }

            public void SyncMats()
            {
                if (_emissiveMaterialData == null || !_initScriptDone || _rebuilding || settings.DisableEmissive)
                    return;

                for (int k = 0; k < _emissiveMaterialData.Count; k++)
                {
                    EmissiveMaterialOptionData data = _emissiveMaterialData[k];
                    for (int i = 0; i < data.MaterialEmissive.Length; i++)
                    {
                        Material m = data.MaterialEmissive[i];
                        if (m == null)
                            continue;

                        Material ms = data.MaterialSource[i];
                        if (ms == null)
                            continue;

                        if (data.ReplaceShaderDirectlyMode)
                            ms.renderQueue = m.renderQueue;
                        else
                            m.renderQueue = ms.renderQueue;

                        SyncMatColor(m, ms, data.Settings[i]);
                    }
                }
            }

            protected static void SyncMatColor(Material mat, Material matOriginalColorSource, WrapTextureSetting setting)
            {
                Color c = Color.white;
                float alphaVal = Math.Max(0f, setting.Alpha * setting.AlphaMultiplier);
                if (alphaVal >= 0f)
                {
                    if (setting.OriginalColor)
                    {
                        if (setting.TextureSettings != null && setting.TextureSettings.EmissiveColor.HasValue)
                        {
                            c = setting.TextureSettings.EmissiveColor.Value;
                        } 
                        else if (matOriginalColorSource != null)
                        {
                            c = matOriginalColorSource.GetColor("_Color");
                            if (c == Color.black)
                            {
                                c = Color.white;
                            }
                        } 
                        else
                        {
                            c = setting.WrapSettings.MainSettings.Color;
                        }
                    }
                    else
                    {
                        c = setting.WrapSettings.MainSettings.Color;
                    }

                    if (!setting.ColorMode)
                    {
                        c.a = alphaVal;
                    }
                    else
                    {
                        c *= alphaVal;
                    }
                    mat.SetColor("_Color", c);
                }
            }

            protected void Build()
            {
                // only build for active atoms
                if (containingAtom == null || !containingAtom.on)
                {
                    if (_verbose)
                        SuperController.LogMessage("EmissiveClothing: Containing atom " + containingAtom.uid + " not on, waiting for it to be on");
                    
                    _waitForOn = true;
                    return;
                }
                try
                {
                    DAZSkinWrap[] allWraps = containingAtom.gameObject.GetComponentsInChildren<DAZSkinWrap>(false);

                    _wraps = new List<DAZSkinWrap>();
                    _ewraps = new List<EmissiveDAZSkinWrap>();

                    _emissiveMaterialData = new List<EmissiveMaterialOptionData>();

                    if (allWraps.Length == 0)
                    {
                        if (_verbose)
                            SuperController.LogMessage("EmissiveClothing: No clothes loaded");
                        return;
                    }

                    // A clothing item already having an EmissiveDAZSkinWrap also have the base DAZSkinWrap.
                    HashSet<string> hasEmissiveWraps = GetEmissiveEnabledWraps(allWraps);

                    // Clothing
                    // \_ Wrap
                    //    |- GPUmaterials (Material[])
                    //    |- DAZSkinWrapMaterialOptions
                    //       |- paramMaterialSlots (int[] - array over slots used in the wrap.GPUmaterials array)

                    bool allSourceBlackColors = true;
                    bool foundDup = false;
                    foreach (DAZSkinWrap wrap in allWraps)
                    {
                        // SuperController.LogMessage("wrap="+wrap.ToString());
                        if (wrap.skin.delayDisplayOneFrame)
                        {
                            if (_verbose)
                                SuperController.LogMessage($"EmissiveClothing: {wrap} is delayed, emissive not set up to handle that, ignoring wrap.");

                            continue;
                        }

                        string wrapName = wrap.ToString();
                        if (wrapName.EndsWith("(" + nameof(EmissiveDAZSkinWrap) + ")"))
                        {
                            // This is already an EmissiveDAZSkinWrap. Ignore and to reduce verbose duplicate emissive logs.
                            continue;
                        }
                        bool isEmissiveDuplicate = hasEmissiveWraps.Contains(wrapName);
                        //if (isEmissiveDuplicate)
                        //{
                        //    // This is expected when having multiple instances of the same script - it's OK
                        //    SuperController.LogMessage($"EmissiveClothing: found dup {wrap}, already handled by another emissive script instance, ignoring wrap.");
                        //    continue;
                        //}

                        HashSet<string> warnedDups = new HashSet<string>();

                        string key = wrap.ToString() + "_" + wrap.wrapName + "_";

                        DAZSkinWrapMaterialOptions[] allOptions = wrap.GetComponents<DAZSkinWrapMaterialOptions>();
                        // bool containsBuiltInEmOption = allOptions.Any(x => IsEmissiveFlagged(x));
                        DAZSkinWrapMaterialOptions[] emissiveOptions = allOptions.Where(mo =>
                        {
                        // too lazy to duplicate all the code for slots2 / simpleMaterial
                        if (mo.paramMaterialSlots?.Length == 0)
                                return false;

                            if (_filterActive)
                            {
                                if (!_filterNames.Contains(key + mo.storeId))
                                    return false;
                            }
                            else
                            {
                                if (!IsEmissiveFlagged(mo))
                                {
                                    return false;
                                }
                            }

                            if (isEmissiveDuplicate)
                            {
                            // This is expected when having multiple instances of the same script - it's OK
                            if (warnedDups.Add(wrapName))
                                {
                                    foundDup = true;
                                    if (_verbose)
                                        SuperController.LogMessage($"EmissiveClothing: found duplicate {wrap}, already handled by another emissive script instance, ignoring wrap.");
                                }
                                return false;
                            }

                            return true;
                        }).ToArray();

                        if (emissiveOptions.Length > 0)
                        {
                            // this wrap has some emissive to do!
                            EmissiveMaterialOptionData data = new EmissiveMaterialOptionData();

                            int length = wrap.GPUmaterials.Length;
                            data.MaterialEmissive = new Material[length];
                            data.MaterialEmissiveBackup = new Material[length];
                            data.MaterialSource = new Material[length];
                            data.MaterialBackup = new Material[length];
                            data.MaterialHidden = new Material[length];
                            data.OriginalDecalTextures = new Texture[length];
                            data.Settings = new WrapTextureSetting[length];
                            data.DoEmissive = new bool[length];
                            data.OriginalWrapDraw = wrap.draw;
                            data.WrapSettings = new WrapSettings(this.settings);

                            foreach (DAZSkinWrapMaterialOptions mo in allOptions)
                            {
                                // too lazy to duplicate all the code for slots2 / simpleMaterial
                                if (mo.paramMaterialSlots?.Length == 0)
                                    continue;

                                bool isSetAsEmissive = emissiveOptions.Contains(mo);

                                foreach (int i in mo.paramMaterialSlots)
                                {
                                    Material mat = wrap.GPUmaterials[i];

                                    WrapTextureSetting wrapTextureSettings = new WrapTextureSetting(data.WrapSettings, null);

                                    data.MaterialSource[i] = mat;
                                    data.DoEmissive[i] = false;
                                    data.Settings[i] = wrapTextureSettings;

                                    if (!isSetAsEmissive)
                                        continue;

                                    // parse settings from decal texture name if available
                                    if (settings.AutomaticTextureSettings)
                                    {
                                        wrapTextureSettings.TextureSettings = TextureSettings.ParseFrom(mo, _verbose);
                                    }
                                    wrapTextureSettings.UpdateSettings();

                                    // set up as an emissive source
                                    Material ourMatEmissive = new Material(_emissiveShader);
                                    Material ourMatBackup = new Material(mat.shader);

                                    ourMatEmissive.CopyPropertiesFromMaterial(mat);
                                    ourMatBackup.CopyPropertiesFromMaterial(mat);

                                    data.MaterialEmissive[i] = ourMatEmissive;
                                    data.MaterialBackup[i] = ourMatBackup;
                                    data.DoEmissive[i] = true;

                                    ourMatEmissive.name = mat.name;
                                    ourMatBackup.name = mat.name;

                                    ourMatEmissive.renderQueue = mat.renderQueue;
                                    ourMatBackup.renderQueue = mat.renderQueue;

                                    SyncMatEmissive(ourMatEmissive, mat, _texPrioritizeOrder);

                                    Material ourMatEmissiveBak = new Material(ourMatEmissive.shader);
                                    ourMatEmissiveBak.CopyPropertiesFromMaterial(ourMatEmissive);
                                    ourMatEmissiveBak.name = ourMatEmissive.name;
                                    ourMatEmissiveBak.renderQueue = ourMatEmissive.renderQueue;
                                    data.MaterialEmissiveBackup[i] = ourMatEmissiveBak;

                                    Color sourceColor = mat.GetColor("_Color");
                                    if (sourceColor.r > 0 || sourceColor.g > 0 || sourceColor.b > 0)
                                    {
                                        allSourceBlackColors = false;
                                    }

                                    SyncMatColor(ourMatEmissive, mat, wrapTextureSettings);

                                    // Looks better for the most part when _DecalTex is not cleared! 
                                    // Let user decide as this is how the original emissive worked.
                                    data.OriginalDecalTextures[i] = mat.GetTexture("_DecalTex");
                                    if (wrapTextureSettings.RenderOriginalWithoutDecal)
                                        mat.SetTexture("_DecalTex", null);

                                    data.MaterialHidden[i] = new Material(mat);
                                    data.MaterialHidden[i].SetFloat("_AlphaAdjust", -1f); // Hidden Toggle Alpha set to 0
                                    data.MaterialHidden[i].renderQueue = mat.renderQueue;

                                    // Activate wrap drawing to see emissive
                                    wrap.draw = true;
                                }
                            }

                            data.Finalize();

                            wrap.BroadcastMessage("OnApplicationFocus", true);
                            // SuperController.LogMessage("wrap.gameObject.AddComponent<EmissiveDAZSkinWrap>");
                            EmissiveDAZSkinWrap ewrap = wrap.gameObject.AddComponent<EmissiveDAZSkinWrap>();
                            ewrap.CopyFrom(wrap, data);
                            // SuperController.LogMessage("wraps.Add(wrap)");

                            _emissiveMaterialData.Add(data);
                            _wraps.Add(wrap);
                            _ewraps.Add(ewrap);

                        }

                    }

                    if (foundDup)
                    {
                        if (_verbose)
                            SuperController.LogMessage("EmissiveClothing: Restoring the scene may result in wrong emissive result (depending on script load order) as duplicates were found. Consider using filters to avoid overlap between emissive clothing items.");
                    }

                    SyncMats();

                    if (_verbose)
                        SuperController.LogMessage("EmissiveClothing: Build done for " + containingAtom.uid);
                } 
                catch (Exception e)
                {
                    SuperController.LogError(e.Message);
                    SuperController.LogError(e.ToString());
                }
            }


            public void Unbuild()
            {
                if (_verbose && _wraps.Count > 0)
                    SuperController.LogMessage("EmissiveClothing: Unbuilding for " + containingAtom.uid);

                for (int i = 0; i < _wraps.Count; i++)
                {
                    try
                    {
                        if (_verbose)
                            SuperController.LogMessage("EmissiveClothing: Unbuilding wrap " + i.ToString());

                        DAZSkinWrap wrap = _wraps[i];
                        EmissiveDAZSkinWrap ewrap = _ewraps[i];

                        ewrap.draw = false;
                        wrap.draw = true;

                        bool destroyed = false;
                        try
                        {
                            // Unity overloads "==" to check true against null on destroyed items.
                            destroyed = (wrap == null || wrap.gameObject == null);
                        }
                        catch (Exception e)
                        {
                            destroyed = true;
                        }

                        if (!destroyed)
                        {
                            DAZSkinWrapControl control = (wrap.gameObject == null ? null : wrap.gameObject.GetComponent<DAZSkinWrapControl>());
                            if (control && (control.wrap == null || control.wrap == ewrap))
                            {
                                control.wrap = wrap;
                            }
                        }
                        else
                        {
                            if (_verbose)
                                SuperController.LogMessage("EmissiveClothing: Wrap destroyed, ignoring restore");
                        }

                        bool fullyRestored = !destroyed;

                        EmissiveMaterialOptionData data = ewrap.Data;
                        if (data != null && !destroyed)
                        {
                            for (int j = 0; j < data.MaterialBackup.Length; j++)
                            {
                                if (data.MaterialBackup[j] == null || !data.DoEmissive[j])
                                    continue;

                                if (data.ReplaceShaderDirectlyMode)
                                {
                                    // swap back to original material source
                                    // Verify that the material is the one we swapped to - and user hasn't done a forced reload while this plugin is active.
                                    // if (ReferenceEquals(wrap.GPUmaterials[j], data.MaterialEmissive[j]))
                                    if (!data.CheckEmissiveMaterialChanged(wrap.GPUmaterials, j) || !wrap.isActiveAndEnabled) // wrap disabled / other clothing preset loaded. Do full restore anyway!
                                    {
                                        // This works as long as the current emissive material has not been changed at all
                                        wrap.GPUmaterials[j] = data.MaterialSource[j];
                                        wrap.GPUmaterials[j].SetTexture("_MainTex", data.MaterialBackup[j].GetTexture("_MainTex"));
                                        wrap.GPUmaterials[j].SetTexture("_AlphaTex", data.MaterialBackup[j].GetTexture("_AlphaTex"));
                                        wrap.GPUmaterials[j].SetTexture("_DecalTex", data.MaterialBackup[j].GetTexture("_DecalTex"));
                                        wrap.GPUmaterials[j].SetFloat("_AlphaAdjust", data.MaterialBackup[j].GetFloat("_AlphaAdjust"));
                                        wrap.GPUmaterials[j].SetColor("_Color", data.MaterialBackup[j].GetColor("_Color"));
                                        wrap.GPUmaterials[j].SetTextureOffset("_MainTex", data.MaterialBackup[j].GetTextureOffset("_MainTex"));
                                        wrap.GPUmaterials[j].SetTextureScale("_MainTex", data.MaterialBackup[j].GetTextureScale("_MainTex"));
                                    }
                                    else
                                    {

                                        if (_verbose)
                                            SuperController.LogMessage("EmissiveClothing: GPUmaterials " + j.ToString() + " changed, doing partial restore in shader emissive mod (sim). Clothing probably forced reload or changed while emissive was active.");

                                        // copy materials back from the current emissive one (as it has the correct properties but wrong shader)
                                        // wrap.GPUmaterials[j] is the one with the emissive shader atm, which has updated materials
                                        // data.MaterialSource[j] is the original material + shader
                                        //Material emissiveMaterial = wrap.GPUmaterials[j];
                                        //Material originalMaterial = data.MaterialSource[j];

                                        //originalMaterial.CopyPropertiesFromMaterial(emissiveMaterial);
                                        //// Copy selected materials
                                        //originalMaterial.SetTexture("_MainTex", emissiveMaterial.GetTexture("_MainTex"));
                                        //originalMaterial.SetTexture("_AlphaTex", emissiveMaterial.GetTexture("_AlphaTex"));
                                        //originalMaterial.SetTexture("_DecalTex", emissiveMaterial.GetTexture("_DecalTex"));
                                        //originalMaterial.SetTexture("_GlossTex", emissiveMaterial.GetTexture("_GlossTex"));
                                        //originalMaterial.SetTexture("_BumpMap", emissiveMaterial.GetTexture("_BumpMap"));
                                        //originalMaterial.SetFloat("_AlphaAdjust", emissiveMaterial.GetFloat("_AlphaAdjust"));
                                        //originalMaterial.SetColor("_Color", emissiveMaterial.GetColor("_Color"));
                                        //originalMaterial.SetTextureOffset("_MainTex", emissiveMaterial.GetTextureOffset("_MainTex"));
                                        //originalMaterial.SetTextureScale("_MainTex", emissiveMaterial.GetTextureScale("_MainTex"));
                                        //wrap.GPUmaterials[j] = originalMaterial;

                                        data.MaterialSource[j].CopyPropertiesFromMaterial(wrap.GPUmaterials[j]);
                                        wrap.GPUmaterials[j] = data.MaterialSource[j];

                                        fullyRestored = false;
                                    }
                                }
                                else
                                {
                                    // Verify that the material is still in use - and user hasn't done a forced reload while this plugin is active.
                                    // if (ReferenceEquals(wrap.GPUmaterials[j], data.MaterialSource[j]))
                                    if (!data.CheckSourceMaterialChanged(wrap.GPUmaterials, j))
                                    {
                                        wrap.GPUmaterials[j].SetTexture("_MainTex", data.MaterialBackup[j].GetTexture("_MainTex"));
                                        wrap.GPUmaterials[j].SetTexture("_AlphaTex", data.MaterialBackup[j].GetTexture("_AlphaTex"));
                                        wrap.GPUmaterials[j].SetTexture("_DecalTex", data.MaterialBackup[j].GetTexture("_DecalTex"));
                                        wrap.GPUmaterials[j].SetFloat("_AlphaAdjust", data.MaterialBackup[j].GetFloat("_AlphaAdjust"));
                                        wrap.GPUmaterials[j].SetColor("_Color", data.MaterialBackup[j].GetColor("_Color"));
                                        wrap.GPUmaterials[j].SetTextureOffset("_MainTex", data.MaterialBackup[j].GetTextureOffset("_MainTex"));
                                        wrap.GPUmaterials[j].SetTextureScale("_MainTex", data.MaterialBackup[j].GetTextureScale("_MainTex"));
                                    }
                                    else
                                    {
                                        if (_verbose)
                                            SuperController.LogMessage("EmissiveClothing: GPUmaterials " + j.ToString() + " changed, ignoring restore. Clothing probably forced reload or changed while emissive was active.");

                                        // wrap.GPUmaterials[j] is the original material + shader already, with already updated materials. Don't restore anything as the user is doing changes.

                                        fullyRestored = false;
                                    }
                                }
                            }
                        }

                        // restore original draw, if everything else was restored as intended (no user changing clothing material textures or force reload).
                        if (fullyRestored)
                        {
                            wrap.draw = data.OriginalWrapDraw;
                        }

                        data.Clear();
                        ewrap.Clear();

                        try
                        {
                            GameObject.Destroy(ewrap);
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError(e.ToString());
                        }
                    }
                    catch (Exception e)
                    {
                        SuperController.LogError(e.ToString());
                    }
                }

                _emissiveMaterialData.Clear();
                _wraps.Clear();
                _ewraps.Clear();
            }

            public bool OutputCurrentSettings()
            {
                if (_ewraps.Count == 0)
                    return false;

                for(int k = 0; k < _ewraps.Count; k++)
                {
                    EmissiveDAZSkinWrap ewrap = _ewraps[k];
                    SuperController.LogMessage("");
                    SuperController.LogMessage("--------------------------------------------------------------------------------------------------------------------------------------------------------------");
                    SuperController.LogMessage("Item wrap: " + ewrap.ToString());
                    var wrap = ewrap.SourceWrap;
                    DAZSkinWrapMaterialOptions[] allOptions = wrap.GetComponents<DAZSkinWrapMaterialOptions>();

                    foreach (DAZSkinWrapMaterialOptions mo in allOptions)
                    {
                        if (mo.paramMaterialSlots?.Length == 0)
                            continue;

                        foreach (int i in mo.paramMaterialSlots)
                        {
                            bool isSetAsEmissive = ewrap.Data.DoEmissive[i];

                            if (!isSetAsEmissive)
                                continue;

                            SuperController.LogMessage("");
                            SuperController.LogMessage("Emissive material: " + mo.storeId);
                            
                            string friendlyName = mo.overrideId;
                            if (friendlyName.StartsWith("+Material"))
                                friendlyName = "Material: " + friendlyName.Substring("+Material".Length);

                            SuperController.LogMessage(friendlyName);

                            string existingDecalText = mo.GetUrlParamValue("customTexture_DecalTex");
                            if (!string.IsNullOrEmpty(existingDecalText))
                            {
                                SuperController.LogMessage("Current decal texture: " + existingDecalText);
                            } 
                            else
                            {
                                SuperController.LogMessage("Current decal texture: <EMPTY>");
                            }
                            
                            string textureName = existingDecalText;

                            string extension = "";
                            int index = textureName.LastIndexOf('.');
                            if (index >= 0)
                            {
                                extension = textureName.Substring(index);
                                textureName = textureName.Substring(0, textureName.Length - extension.Length);
                            }

                            if (!string.IsNullOrEmpty(textureName))
                            {
                                // Cut any paths
                                index = textureName.LastIndexOfAny(new char[] { '/', '\\', ':' });
                                if (index >= 0)
                                {
                                    textureName = textureName.Substring(index + 1);
                                }

                                // Parse existing automatic settings from texture and remove that part (if it exist)
                                // Any part after "(em)" will be removed
                                index = textureName.IndexOf("(em)", 0, StringComparison.InvariantCultureIgnoreCase);
                                if (index >= 0)
                                {
                                    textureName = textureName.Substring(0, index);
                                    // skip any ending "_", but only 1 occurance
                                    if (textureName.Length > 1 && textureName.EndsWith("_"))
                                    {
                                        textureName = textureName.Substring(0, textureName.Length - 1);
                                    }
                                }
                            }

                            string currentAutomaticSettings = "_(em)";
                            if (!settings.OriginalColor)
                            {
                                currentAutomaticSettings += "_c" + ColorToString(settings.Color);
                            }
                            currentAutomaticSettings += "_a" + ((int)(settings.Alpha * 100f)).ToString();
                            currentAutomaticSettings += "_m" + (settings.ColorMode ? "1" : "0");
                            currentAutomaticSettings += "_s" + (settings.ShaderMode ? "1" : "0");
                            currentAutomaticSettings += "_r" + (settings.RenderOriginal ? "1" : "0");
                            currentAutomaticSettings += "_d" + (settings.RenderOriginalWithoutDecal ? "1" : "0");
                            currentAutomaticSettings += "_";

                            SuperController.LogMessage("Settings to apply: " + currentAutomaticSettings);

                            // Insert currentAutomaticSettings at the end of the texture name
                            string newTextureName = textureName + currentAutomaticSettings + extension;
                            SuperController.LogMessage("New decal texture name for automatic settings: " + newTextureName);
                        }
                    }
                }
                return true;
            }

        }
        #endregion

        #region EMISSIVE_WRAP
        public class TextureSettings
        {
            // This will contain settings from TEXTURE, parsed from texture name!
            public Color? EmissiveColor;
            public float? Alpha;
            public bool? ColorMode;
            public bool? ShaderMode;
            public bool? RenderOriginal;
            public bool? RenderOriginalWithoutDecal;

            public static TextureSettings ParseFrom(DAZSkinWrapMaterialOptions mo, bool verbose)
            {
                TextureSettings textureSettings = new TextureSettings();
                
                string existingDecalText = mo.GetUrlParamValue("customTexture_DecalTex");
                if (string.IsNullOrEmpty(existingDecalText))
                    return textureSettings;

                // find "(em)" within the filename
                existingDecalText = existingDecalText.ToLowerInvariant();
                int index = existingDecalText.LastIndexOf("(em)");
                if (index <= -1)
                    return textureSettings;
                
                // contains (em) parsable settings
                string settings = existingDecalText.Substring(index + "(em)".Length);
                
                // remove file extension
                index = settings.LastIndexOf(".");
                if (index >= 0)
                {
                    settings = settings.Substring(0, index);
                }

                if (verbose)
                    SuperController.LogMessage("EmissiveClothing: Automatic settings: " + settings + " for " + mo.overrideId);

                // settings now contains the "(em)_X_Y_Z_" settings part
                string[] parts = settings.Split(new char[] { '_' });

                for(int i = 0; i < parts.Length; i ++)
                {
                    string part = parts[i];

                    if (part == "d1")
                        textureSettings.RenderOriginalWithoutDecal = true;
                    else if (part == "d0")
                        textureSettings.RenderOriginalWithoutDecal = false;

                    if (part == "r1")
                        textureSettings.RenderOriginal = true;
                    else if (part == "r0")
                        textureSettings.RenderOriginal = false;

                    if (part == "s1")
                        textureSettings.ShaderMode = true;
                    else if (part == "s0")
                        textureSettings.ShaderMode = false;

                    if (part == "m1")
                        textureSettings.ColorMode = true;
                    else if (part == "m0")
                        textureSettings.ColorMode = false;
                    
                    if (part.StartsWith("a"))
                    {
                        // alpha
                        int a;
                        if (int.TryParse(part.Substring(1), out a))
                        {
                            textureSettings.Alpha = (float)a / 100f;
                        }
                    }

                    if (part.StartsWith("c") && part.Length == 7)
                    {
                        Color c;
                        if (TryParseColor(part.Substring(1), out c))
                        {
                            textureSettings.EmissiveColor = c;
                        }
                    }
                }

                return textureSettings;
            }
        }

        public class WrapSettings
        {
            // This will contain settings for MATERIAL, a combination of TEXTURE (where settings are applicable), the WRAP itself and PLUGIN INPUT
            public IEmissiveSettings MainSettings;

            // RenderOriginal is currently at CLOTHING level, not MATERIAL level.
            // ReplaceShader is currently at CLOTHING level, not MATERIAL level.
            // Has to do with UpdateVertsGPU and DrawMeshGPU as it's done on a per Wrap level, not material/shader level.

            // wrap level
            public bool? RenderOriginalOverride;

            // wrap level
            public bool? ShaderModeOverride;

            // wrap level
            public bool ShaderMode
            {
                get
                {
                    if (ShaderModeOverride.HasValue)
                        return ShaderModeOverride.Value;

                    return MainSettings.ShaderMode;
                }
            }

            // wrap level
            public bool RenderOriginal
            {
                get
                {
                    if (RenderOriginalOverride.HasValue)
                        return RenderOriginalOverride.Value;

                    return MainSettings.RenderOriginal;
                }
            }

            public WrapSettings(IEmissiveSettings mainSettings)
            {
                MainSettings = mainSettings;
            }
        }

        public class WrapTextureSetting
        {
            public WrapSettings WrapSettings;
            public TextureSettings TextureSettings;

            public WrapTextureSetting(WrapSettings wrap, TextureSettings texture)
            {
                WrapSettings = wrap;
                TextureSettings = texture;
            }

            // plugin global
            public float AlphaMultiplier
            {
                get
                {
                    return WrapSettings.MainSettings.AlphaMultiplier;
                }
            }

            // plugin global
            public bool OriginalColor
            {
                get
                {
                    return WrapSettings.MainSettings.OriginalColor;
                }
            }

            // materialOption level
            public Color EmissiveColor
            {
                get
                {
                    if (TextureSettings != null && TextureSettings.EmissiveColor.HasValue)
                        return TextureSettings.EmissiveColor.Value;

                    return WrapSettings.MainSettings.Color;
                }
            }

            // materialOption level
            public bool ColorMode 
            { 
                get 
                {
                    if (TextureSettings != null && TextureSettings.ColorMode.HasValue)
                        return TextureSettings.ColorMode.Value;

                    return WrapSettings.MainSettings.ColorMode;
                } 
            }

            // materialOption level
            public float Alpha
            {
                get
                {
                    if (TextureSettings != null && TextureSettings.Alpha.HasValue)
                        return TextureSettings.Alpha.Value;

                    return WrapSettings.MainSettings.Alpha;
                }
            }            

            // materialOption level
            public bool RenderOriginalWithoutDecal
            {
                get
                {
                    if (TextureSettings != null && TextureSettings.RenderOriginalWithoutDecal.HasValue)
                        return TextureSettings.RenderOriginalWithoutDecal.Value;

                    return WrapSettings.MainSettings.RenderOriginalWithoutDecal;
                }
            }

            public void UpdateSettings()
            {
                if (TextureSettings == null)
                    return;

                if (!WrapSettings.ShaderModeOverride.HasValue && TextureSettings.ShaderMode.HasValue)
                    WrapSettings.ShaderModeOverride = TextureSettings.ShaderMode.Value;

                if (!WrapSettings.RenderOriginalOverride.HasValue && TextureSettings.RenderOriginal.HasValue)
                    WrapSettings.RenderOriginalOverride = TextureSettings.RenderOriginal.Value;
            }
        }

        public class EmissiveMaterialOptionData
        {
            // !All arrays here are synced in length and material index!

            // The emissive materials ("our material")
            public Material[] MaterialEmissive;
            // Material emissive backup, as it was created.
            public Material[] MaterialEmissiveBackup;
            // Stores the source materials
            public Material[] MaterialSource;
            // Stores some original data from the source that needs to be restored during unbuild.
            public Material[] MaterialBackup;
            // Hidden material to be emissive
            public Material[] MaterialHidden;
            // Keeps a reference to the original source decal texture (if there was any)
            public Texture[] OriginalDecalTextures;
            // Material/Texture setting used for this material
            public WrapTextureSetting[] Settings;
            // Keeps track on which materials should be emissive and not
            public bool[] DoEmissive;

            public Material[] RenderOriginalTrueMaterials;
            public Material[] RenderOriginalFalseMaterials;
            public Material[] ProcessMesh;

            // If the shader in the native wrap is replaced (which has a few more benifits and features) (applies to all GPUMaterials)
            public bool ReplaceShaderDirectlyMode = false;

            // If render original is active (applies to all GPUMaterials)
            public bool RenderOriginalMode = false;

            // Tracks if the original wrap was drawn or not
            public bool OriginalWrapDraw = true;

            public WrapSettings WrapSettings;

            public void Clear()
            {
                MaterialEmissive = null;
                MaterialEmissiveBackup = null;
                MaterialSource = null;
                MaterialBackup = null;
                MaterialHidden = null;
                DoEmissive = null;
                RenderOriginalTrueMaterials = null;
                RenderOriginalFalseMaterials = null;
                ProcessMesh = null;
                OriginalDecalTextures = null;
                Settings = null;
                if (WrapSettings != null)
                {
                    WrapSettings.MainSettings = null;
                    WrapSettings = null;
                }
            }
            
            public bool CheckEmissiveMaterialChanged(Material[] currentGPUMaterials, int j)
            {
                // Verify that the material is still in use - and user hasn't done a forced reload while this plugin is active.
                if (ReferenceEquals(currentGPUMaterials[j], MaterialEmissive[j]))
                {
                    // check textures relevant to emissive operations
                    if (MaterialEmissive[j].GetTexture("_MainTex") != MaterialEmissiveBackup[j].GetTexture("_MainTex") ||
                        MaterialEmissive[j].GetTexture("_AlphaTex") != MaterialEmissiveBackup[j].GetTexture("_AlphaTex") ||
                        MaterialEmissive[j].GetTexture("_DecalTex") != MaterialEmissiveBackup[j].GetTexture("_DecalTex"))
                    {
                        return true;
                    }
                }
                else
                {
                    // materials no longer matching
                    return true;
                }
                // nothing changed
                return false;
            }
            
            public bool CheckSourceMaterialChanged(Material[] currentGPUMaterials, int j)
            {
                // Verify that the material is still in use - and user hasn't done a forced reload while this plugin is active.
                if (ReferenceEquals(currentGPUMaterials[j], MaterialSource[j]))
                {
                    // check textures relevant to emissive operations
                    if (MaterialSource[j].GetTexture("_MainTex") != MaterialBackup[j].GetTexture("_MainTex") ||
                        MaterialSource[j].GetTexture("_AlphaTex") != MaterialBackup[j].GetTexture("_AlphaTex") ||
                        MaterialSource[j].GetTexture("_DecalTex") != null && MaterialSource[j].GetTexture("_DecalTex") != OriginalDecalTextures[j]) // DecalTex might be null by user choice! Don't consider a null _DecalTex as a change!
                    {
                        // textures no longer matching
                        return true;
                    }
                }
                else
                {
                    // materials no longer matching
                    return true;
                }
                // nothing changed
                return false;
            }

            public void Finalize()
            {
                // Should the emissive be the main material in VAM.
                ReplaceShaderDirectlyMode = WrapSettings.ShaderMode;
                RenderOriginalMode = WrapSettings.RenderOriginal;

                RenderOriginalTrueMaterials = new Material[MaterialEmissive.Length];
                RenderOriginalFalseMaterials = new Material[MaterialEmissive.Length];
                ProcessMesh = new Material[MaterialEmissive.Length];

                for (int i = 0; i < MaterialEmissive.Length; i++)
                {
                    RenderOriginalFalseMaterials[i] = MaterialHidden[i];
                    RenderOriginalTrueMaterials[i] = MaterialSource[i];

                    if (DoEmissive[i])
                    {
                        ProcessMesh[i] = MaterialEmissive[i];
                    } 
                    else
                    {
                        if (ReplaceShaderDirectlyMode)
                        {
                            // the shader in the native wrap will handle the rendering for non-emissive materials
                            RenderOriginalTrueMaterials[i] = null;
                            RenderOriginalFalseMaterials[i] = null;
                            ProcessMesh[i] = null;
                        } 
                        else
                        {
                            // The hidden material will be rendered with the source material
                            // creating a native rendered result for non-emissive materials in a clothing item
                            RenderOriginalTrueMaterials[i] = MaterialSource[i];
                            RenderOriginalFalseMaterials[i] = MaterialSource[i];
                            ProcessMesh[i] = MaterialSource[i];
                        }
                    }
                }
            }
        }

        // The only thing we absolutely need and can't get normally from 
        // DAZSkinWrap is the ComputeBuffers; Material.GetBuffer does not exist
        public class EmissiveDAZSkinWrap : DAZSkinWrap
        {
            Dictionary<int, ComputeBuffer> _emissiveMaterialVertsBuffers = new Dictionary<int, ComputeBuffer>();
            Dictionary<int, ComputeBuffer> _emissiveMaterialNormalsBuffers = new Dictionary<int, ComputeBuffer>();
            Dictionary<int, ComputeBuffer> _emissiveMaterialTangentsBuffers = new Dictionary<int, ComputeBuffer>();

            bool _clearDone;

            public DAZSkinWrap SourceWrap;

            public EmissiveMaterialOptionData Data;

            public void CopyFrom(DAZSkinWrap wrap, EmissiveMaterialOptionData data)
            {
                base.skinTransform = wrap.skinTransform;
                base.skin = wrap.skin;
                base.dazMesh = wrap.dazMesh;
                base.GPUSkinWrapper = wrap.GPUSkinWrapper;
                base.GPUMeshCompute = wrap.GPUMeshCompute;
                base.GPUmaterials = wrap.GPUmaterials;
                base.morphCopyFromTransform = wrap.morphCopyFromTransform;
                base.morphCopyFrom = wrap.morphCopyFrom;
                base.CopyMaterials();

                Data = data;

                this.SourceWrap = wrap;

                base.wrapName = wrap.wrapName;
                base.wrapStore = wrap.wrapStore;

                base.surfaceOffset = wrap.surfaceOffset;
                base.defaultSurfaceOffset = wrap.defaultSurfaceOffset;
                base.additionalThicknessMultiplier = wrap.additionalThicknessMultiplier;
                base.defaultAdditionalThicknessMultiplier = wrap.defaultAdditionalThicknessMultiplier;

                if (data.ReplaceShaderDirectlyMode)
                {
                    base.draw = false;
                    // Swap the materials from the original source, works out much better!
                    for (int i = 0; i < data.DoEmissive.Length; i++)
                    {
                        if (data.DoEmissive[i])
                        {
                            wrap.GPUmaterials[i] = data.MaterialEmissive[i];
                        }
                    }
                }
                else
                {
                    wrap.draw = false;
                    base.draw = true;

                    var control = GetComponent<DAZSkinWrapControl>();
                    if (control && (control.wrap == null || control.wrap == wrap))
                    {
                        control.wrap = this;
                    }
                }
                
            }


            void LateUpdate()
            {
                if (!_clearDone) // (Data.RenderOriginalMode != _clearDone)
                {
                    materialVertsBuffers.Clear();
                    materialNormalsBuffers.Clear();
                    materialTangentsBuffers.Clear();
                    _clearDone = true; // Data.RenderOriginalMode;
                }

                if (Data.ReplaceShaderDirectlyMode)
                {
                    if (Data.RenderOriginalMode)
                    {
                        // Render original in ReplaceShaderDirectlyMode mode is a bit hacky...

                        // Grab the magic buffer data from the ProcessMesh materials populated within VAM already
                        // copy it into the original shader (which isn't the active in ReplaceShaderDirectlyMode).
                        // Now we can render the correct mesh.
                        // However, emissive have some material specific overrides so that needs to be restored too for a correct look!
                        for (int i = 0; i < Data.MaterialSource.Length; i++)
                        {
                            if (Data.DoEmissive[i])
                            {
                                Data.RenderOriginalTrueMaterials[i].CopyPropertiesFromMaterial(Data.ProcessMesh[i]);
                                Data.RenderOriginalTrueMaterials[i].SetTexture("_MainTex", Data.MaterialBackup[i].GetTexture("_MainTex"));
                                Data.RenderOriginalTrueMaterials[i].SetTexture("_DecalTex", Data.MaterialBackup[i].GetTexture("_DecalTex"));
                                Data.RenderOriginalTrueMaterials[i].SetFloat("_AlphaAdjust", Data.MaterialBackup[i].GetFloat("_AlphaAdjust"));
                                Data.RenderOriginalTrueMaterials[i].SetColor("_Color", Data.MaterialBackup[i].GetColor("_Color"));
                                Data.RenderOriginalTrueMaterials[i].SetTextureOffset("_MainTex", Data.MaterialBackup[i].GetTextureOffset("_MainTex"));
                                Data.RenderOriginalTrueMaterials[i].SetTextureScale("_MainTex", Data.MaterialBackup[i].GetTextureScale("_MainTex"));
                            }
                        }

                        Material[] tempGPUMaterials = base.GPUmaterials;

                        base.GPUmaterials = Data.RenderOriginalTrueMaterials;
                        UpdateVertsGPU();
                        DrawMeshGPU();
                        
                        base.GPUmaterials = tempGPUMaterials;
                    }
                    return;
                }

                {
                    Material[] tempGPUMaterials = base.GPUmaterials;
                    if (!Data.RenderOriginalMode)
                    {
                        base.GPUmaterials = Data.RenderOriginalFalseMaterials;
                        UpdateVertsGPU();
                        DrawMeshGPU();
                    }
                    else
                    {
                        base.GPUmaterials = Data.RenderOriginalTrueMaterials;
                        UpdateVertsGPU();
                        DrawMeshGPU();
                    }
                    base.GPUmaterials = tempGPUMaterials;
                }

                for (int i = 0; i < Data.ProcessMesh.Length; i++)
                {
                    if (Data.ProcessMesh[i] == null || !Data.DoEmissive[i])
                        continue;

                    Data.ProcessMesh[i].renderQueue = Data.MaterialSource[i].renderQueue;

                    // SuperController.LogMessage("_ClothVertices HasBuffer: " + Data.MaterialSource[i].HasBuffer("_ClothVertices").ToString());

                    UpdateBuffer(Data.ProcessMesh, _emissiveMaterialVertsBuffers, base.materialVertsBuffers, "verts", i);
                    //UpdateBuffer(Data.ProcessMesh, _emissiveMaterialNormalsBuffers, base.materialNormalsBuffers, "normals", i);
                    //UpdateBuffer(Data.ProcessMesh, _emissiveMaterialTangentsBuffers, base.materialTangentsBuffers, "tangents", i);

                    //if (Data.DoEmissive[i])
                    //{    
                    Graphics.DrawMesh(base.mesh, Matrix4x4.identity, Data.ProcessMesh[i], 0, null, i, null, false, false);
                    //} 
                    //else
                    //{
                        // Original mesh and material already drawn, don't draw again
                        // Graphics.DrawMesh(base.mesh, Matrix4x4.identity, Data.ProcessMesh[i], 0, null, i, null, dazMesh.castShadows, dazMesh.receiveShadows);
                    //}
                    
                }
            }

            private void UpdateBuffer(Material[] material, Dictionary<int, ComputeBuffer> ours, Dictionary<int, ComputeBuffer> theirs, string buf, int i)
            {
                ComputeBuffer ourB, theirB;
                ours.TryGetValue(i, out ourB);
                theirs.TryGetValue(i, out theirB);
                if (ourB != theirB) {
                    material[i].SetBuffer(buf, theirB);
                    ours[i] = theirB;
                }
            }

            //public void OnDestroy()
            //{
            //    base.OnDestroy();
            //    SuperController.LogMessage("Destroy test");
            //}

            public void Clear()
            {
                SourceWrap = null;
                Data = null;
                materialVertsBuffers.Clear();
                materialNormalsBuffers.Clear();
                materialTangentsBuffers.Clear();
                _emissiveMaterialVertsBuffers.Clear();
                _emissiveMaterialNormalsBuffers.Clear();
                _emissiveMaterialTangentsBuffers.Clear();
            }

        }
        #endregion

        #region STATIC_HELPERS
        public static void SetTextFieldHeight(UIDynamicTextField textField, int height)
        {
            LayoutElement component = textField.GetComponent<LayoutElement>();
            if (component != null)
            {
                component.minHeight = height;
                component.preferredHeight = height;
            }
            textField.height = height;
        }

        public static bool IsVAMLoading(bool missingLoadingIconEqualsLoading = true)
        {
            if (SuperController.singleton.isLoading)
                return true;

            Transform loadingIcon = SuperController.singleton.loadingIcon;

            if (loadingIcon == null || loadingIcon.gameObject == null)
                return missingLoadingIconEqualsLoading;

            if (loadingIcon.gameObject.active)
                return true;

            return false;
        }

        public static string GetClothingName(string storeId)
        {
            storeId = storeId.Replace("(" + nameof(DAZSkinWrap) + ")", "");

            // remove author to shorten the label, the user will understand which clothing item they are working on
            int index = storeId.IndexOf(":");
            if (index > 0)
            {
                storeId = storeId.Substring(index + 1);
            }
            // Add some space around the last "Material" word
            index = storeId.LastIndexOf("Material");
            if (index > 0)
            {
                storeId = storeId.Substring(0, index);
            }
            return storeId;
        }

        public static string FancifyMaterialOptionLabel(string wrapName, string label)
        {

            label = label.Replace("(" + nameof(DAZSkinWrap) + ")", "");

            // remove author to shorten the label, the user will understand which clothing item they are working on
            int index = label.IndexOf(":");
            if (index > 0)
            {
                label = label.Substring(index + 1);
            }

            // Add some space around the last "Material" word
            index = label.LastIndexOf("Material");
            if (index > 0)
            {
                label = label.Insert(index + "Material".Length, " ");
                label = label.Insert(index, "\n");
            }

            if (!string.IsNullOrEmpty(wrapName) && wrapName != "Normal")
            {
                index = label.IndexOf("\n");
                if (index > 0)
                {
                    label = label.Insert(index, $" ({wrapName})");
                }
                else
                {
                    label = $"{label} {wrapName}";
                }
            }

            return label;
        }

        /// <summary>
        /// Returns the path to the BUNDLE_NAME
        /// </summary>
        public static string GetPluginBundlePath(MVRScript script)
        {
            // SuperController.LogMessage("GetPluginPath");

            // string loadDir = SuperController.singleton.currentLoadDir;
            // SuperController.LogMessage(SuperController.singleton.currentSaveDir);
            // SuperController.LogMessage(script.storeId); // the id of the plugin as generated by VAM
            string pluginId = script.storeId.Split('_')[0]; // first part is the plugin ID in the MVRPluginManager.
            MVRPluginManager man = (script.manager != null ? script.manager : (script.containingAtom.GetStorableByID("PluginManager") as MVRPluginManager));
            string pathToScriptFile = man.GetJSON(true, true)["plugins"][pluginId].Value; // the path to the plugin file source
            // SuperController.LogMessage(pathToScriptFile); // "Custom/Scripts/FolderA/ScriptFile.cs/cslist"

            if (pathToScriptFile.Contains(":/"))
            {
                // var packaged script
                string varPackageName = pathToScriptFile.Substring(0, pathToScriptFile.IndexOf(":/"));
                return $"{varPackageName}:/Custom/Assets/{BUNDLE_NAME}";
            }
            // non var packaged script
            // simply require the emissiveshader.assetbundle to be in Custom/Assets/emissiveshader.assetbundle to let VAM only load 1 instance of this assetbundle
            return $"Custom/Assets/{BUNDLE_NAME}";

            /*
            // If the emissiveshader.assetbundle should be required to be next to the file itself:
            string pathToScriptFolder = "";
            if (pathToScriptFile.Contains("/") || pathToScriptFile.Contains("\\")) { pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' })); }
            if (pathToScriptFolder.StartsWith(".")) { pathToScriptFolder = pathToScriptFolder.Substring(1); }
            if (pathToScriptFolder.StartsWith("/")) { pathToScriptFolder = pathToScriptFolder.Substring(1); }
            // SuperController.LogMessage(pathToScriptFolder); // "Scripts/FolderA/PluginFolder" OR "AddonPackageVarFile.#:/Scripts/FolderA/PluginFolder" for var-packages.
            return pathToScriptFolder + "/emissiveshader.assetbundle";
            */
        }

        public static void SyncMatEmissive(Material matEmissive, Material matSource, string texSource)
        {
            // Grab decal, if not available use the alpha texture as backup
            string sourceTex = "_DecalTex";
            var eTex = matSource.GetTexture(sourceTex);

            switch (texSource)
            {
                case TEX_SOURCE_DECALTEX:
                    if (eTex == null && matSource.GetTexture("_AlphaTex") != null)
                    {
                        sourceTex = "_AlphaTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_ALPHATEX:
                    if (matSource.GetTexture("_AlphaTex") != null)
                    {
                        sourceTex = "_AlphaTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_SPECTEX:
                    if (matSource.GetTexture("_SpecTex") != null)
                    {
                        sourceTex = "_SpecTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_GLOSSTEX:
                    if (matSource.GetTexture("_GlossTex") != null)
                    {
                        sourceTex = "_GlossTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_BUMPMAP:
                    if (matSource.GetTexture("_BumpMap") != null)
                    {
                        sourceTex = "_BumpMap";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
                case TEX_SOURCE_MAINTEX:
                    if (matSource.GetTexture("_MainTex") != null)
                    {
                        sourceTex = "_MainTex";
                        eTex = matSource.GetTexture(sourceTex);
                    }
                    break;
            }

            matEmissive.SetTexture("_MainTex", eTex);
            matEmissive.SetColor("_Color", matSource.GetColor("_Color"));
            matEmissive.SetFloat("_AlphaAdjust", matSource.GetFloat("_AlphaAdjust"));

            // As the offset and scale is from the emissive texture source too, these properties must ALSO be restored throughout the logic!
            matEmissive.SetTextureOffset("_MainTex", matSource.GetTextureOffset(sourceTex));
            matEmissive.SetTextureScale("_MainTex", matSource.GetTextureScale(sourceTex));
        }

        public static IEnumerator WaitLoading(bool verbose)
        {
            yield return new WaitForEndOfFrame();

            bool loadMessageShown = false;
            DateTime lastLoading = DateTime.Now;
            int loadingfreeFrames = 0;
            while (true)
            {
                if (IsVAMLoading())
                {
                    lastLoading = DateTime.Now;
                    loadingfreeFrames = 0;
                    if (!loadMessageShown)
                    {
                        loadMessageShown = true;
                        if (verbose)
                            SuperController.LogMessage("EmissiveClothing: Waiting for loading to complete");
                    }

                    yield return new WaitForEndOfFrame();
                }
                else
                {
                    // debounce loading wait, there can be a frame or two where the loading is "paused" before activating again.
                    // wait for at least 0.4 seconds of non-loading before continuing and AT LEAST 6 frames, this makes session mode of this plugin work must more stable together with non-session instances of this plugin.
                    yield return new WaitForEndOfFrame();

                    loadingfreeFrames++;
                    TimeSpan ts = DateTime.Now.Subtract(lastLoading);
                    if (!IsVAMLoading() && ts.TotalSeconds > 0.4 && loadingfreeFrames >= 6)
                        break;
                }
            }

            yield return new WaitForEndOfFrame();
        }

        public static HashSet<string> GetEmissiveEnabledWraps(DAZSkinWrap[] wraps)
        {
            return new HashSet<string>(wraps.Select(x => x.ToString()).
                                             Where(x => x.EndsWith("(" + nameof(EmissiveDAZSkinWrap) + ")")).
                                             SelectMany(x => {
                                                 // Return both "ClothingName (EmissiveDAZSkinWrap)" and "ClothingName (DAZSkinWrap)" as they have the emissive effect active
                                                 return new string[] {
                                                     x,
                                                     x.Substring(0, x.Length - ("(" + nameof(EmissiveDAZSkinWrap) + ")").Length) + "(" + nameof(DAZSkinWrap) + ")" };
                                             }));
        }

        public static bool ExistEMWrapWithoutEmissiveActive(Atom atom, bool verbose)
        {
            foreach (DAZSkinWrap wrap in WrapsWithoutEmissiveDAZSkinWrap(atom))
            {
                DAZSkinWrapMaterialOptions[] allOptions = wrap.GetComponents<DAZSkinWrapMaterialOptions>();
                bool containsBuiltInEmOption = allOptions.Any(x => IsEmissiveFlagged(x));
                if (containsBuiltInEmOption)
                {
                    if (verbose)
                        SuperController.LogMessage("EmissiveClothing: Wrap " + wrap.ToString() + " contains (em) without emissive active yet");

                    return true;
                }
            }
            return false;
        }

        public static IEnumerable<DAZSkinWrap> WrapsWithoutEmissiveDAZSkinWrap(Atom atom)
        {
            DAZSkinWrap[] allWraps = atom.gameObject.GetComponentsInChildren<DAZSkinWrap>(false);
            HashSet<string> hasEmissiveWraps = GetEmissiveEnabledWraps(allWraps);

            foreach (DAZSkinWrap wrap in allWraps)
            {
                if (wrap.skin.delayDisplayOneFrame)
                    continue;

                if (hasEmissiveWraps.Contains(wrap.ToString()))
                    continue;

                yield return wrap;
            }
        }

        public static bool CheckWrapMaterialsChanged(DAZSkinWrap wrap, EmissiveDAZSkinWrap ewrap)
        {
            bool destroyed = false;
            try
            {
                // Unity overloads "==" to check true against null on destroyed items.
                destroyed = (wrap == null || wrap.gameObject == null);
            }
            catch (Exception e)
            {
                destroyed = true;
            }
            if (destroyed)
            {
                return true;
            }

            EmissiveMaterialOptionData data = ewrap.Data;
            if (data != null)
            {
                for (int j = 0; j < data.MaterialBackup.Length; j++)
                {
                    if (data.MaterialBackup[j] == null || !data.DoEmissive[j])
                        continue;

                    if (data.ReplaceShaderDirectlyMode)
                    {
                        // MaterialEmissive is the active material containing the original textures if not modified
                        if (data.CheckEmissiveMaterialChanged(wrap.GPUmaterials, j))
                        {
                            return true;
                        }
                    }
                    else
                    {
                        // MaterialSource is the original and still the active material
                        if (data.CheckSourceMaterialChanged(wrap.GPUmaterials, j))
                        {
                            return true;
                        }
                    }
                }

            }

            return false;
        }

        public static bool IsEmissiveFlagged(DAZSkinWrapMaterialOptions mo)
        {
            if (mo.overrideId.Contains("(em)"))
                return true;

            string existingDecalText = mo.GetUrlParamValue("customTexture_DecalTex");
            if (!string.IsNullOrEmpty(existingDecalText))
            {
                // find "(em)" within the filename
                int index = existingDecalText.LastIndexOfAny(new char[] { '/', '\\', ':' });
                if (index >= 0)
                {
                    index = existingDecalText.IndexOf("(em)", index, StringComparison.InvariantCultureIgnoreCase);
                } 
                else
                {   
                    index = existingDecalText.IndexOf("(em)", 0, StringComparison.InvariantCultureIgnoreCase);
                }
                return (index >= 0);
            }

            return false;
        }

        public static string ColorToString(Color c)
        {
            byte[] rgb = new byte[] { (byte)(c.r * 255f), (byte)(c.g * 255f), (byte)(c.b * 255f) };
            string hex = BitConverter.ToString(rgb).Replace("-", string.Empty).ToLowerInvariant();
            return hex;
        }
        public static bool TryParseColor(string s, out Color color)
        {
            color = Color.white;
            if (s.Length != 6)
                return false;

            byte r, g, b;
            if (byte.TryParse(s.Substring(0, 2), System.Globalization.NumberStyles.HexNumber, System.Globalization.CultureInfo.InvariantCulture, out r) &&
                byte.TryParse(s.Substring(2, 2), System.Globalization.NumberStyles.HexNumber, System.Globalization.CultureInfo.InvariantCulture, out g) &&
                byte.TryParse(s.Substring(4, 2), System.Globalization.NumberStyles.HexNumber, System.Globalization.CultureInfo.InvariantCulture, out b))
            {
                color = new Color((float)r / 255f, (float)g / 255f, (float)b / 255f, 1f);
                return true;
            }
            return false;
        }

        /// <summary>
        /// The pluginUid data goes STALE when plugin presets are loaded! Might even go stale if adding or removing too!
        /// Read this again before analyzing plugin data.
        /// </summary>
        public static bool ContainsScript(Atom atom, string plugin)
        {
            if (atom == null)
            {
                return false;
            }

            MVRPluginManager pluginManager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            if (pluginManager == null)
            {
                return false;
            }

            // list existing plugins
            foreach (string storeableId in atom.GetStorableIDs())
            {
                if (storeableId.StartsWith("plugin#"))
                {
                    if (storeableId.EndsWith("_" + plugin))
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        //public static void Test(DAZSkinWrapMaterialOptions mo)
        //{
        //    // DAZSkinWrapMaterialOptions : MaterialOptions
        //    //SuperController.LogMessage(mo.startingTextureGroup1Set);
        //    //SuperController.LogMessage(mo.currentTextureGroup1Set);
        //    //SuperController.LogMessage(mo.startingTextureGroup2Set);
        //    //SuperController.LogMessage(mo.currentTextureGroup2Set);
        //    //SuperController.LogMessage(mo.startingTextureGroup3Set);
        //    //SuperController.LogMessage(mo.currentTextureGroup3Set);
        //    //SuperController.LogMessage(mo.startingTextureGroup4Set);
        //    //SuperController.LogMessage(mo.currentTextureGroup4Set);
        //    //SuperController.LogMessage(mo.startingTextureGroup5Set);
        //    //SuperController.LogMessage(mo.currentTextureGroup5Set);
        //    SuperController.LogMessage(mo.storeId);
        //    // SuperController.LogMessage(mo.GetStringParamValue("customTexture_MainTex"));
        //    foreach (string str in mo.GetUrlParamNames())
        //    {
        //        SuperController.LogMessage(str);
        //        SuperController.LogMessage(mo.GetUrlParamValue(str)); // SCORE!
        //    }
        //    //JSONStorable store = mo.GetStorableById(mo.storeId);
        //    //if (store != null)
        //    //{

        //    //}
        //    //SuperController.LogMessage(mo.GetJSON(true, true, true).ToString());
        //    //Test2(mo.textureGroup1);
        //    //Test2(mo.textureGroup2);
        //    //Test2(mo.textureGroup3);
        //    //Test2(mo.textureGroup4);
        //    //Test2(mo.textureGroup5);
        //}
        //public static void Test2(MaterialOptionTextureGroup tg)
        //{
        //    // SuperController.LogMessage(tg.fourthTextureName);
        //    foreach (var set in tg.sets)
        //    {
        //        SuperController.LogMessage(set.name);
        //    }
        //}
        #endregion

    }

}
